#include "opendrive_road_node.h"
#include "opendrive_road.h"
#include "exceptions.h"
#include <math.h>

COpendriveRoadNode::COpendriveRoadNode()
{
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->start_point.x=0.0;
  this->start_point.y=0.0;
  this->start_point.heading=0.0;
  this->lanes.clear();
  this->geometries.clear();
  this->index=-1;
  this->parent_segment=NULL;
}

COpendriveRoadNode::COpendriveRoadNode(const COpendriveRoadNode &object)
{
  COpendriveLink *new_link;

  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
  this->start_point.x=object.start_point.x;
  this->start_point.y=object.start_point.y;
  this->start_point.heading=object.start_point.heading;
  this->lanes.resize(object.lanes.size());
  for(unsigned int i=0;i<object.lanes.size();i++)
    this->lanes[i]=object.lanes[i];
  this->geometries.resize(object.geometries.size());
  for(unsigned int i=0;i<object.geometries.size();i++)
    this->geometries[i]=object.geometries[i]->clone();
  this->links.clear();
  for(unsigned int i=0;i<object.links.size();i++)
  {
    new_link=new COpendriveLink(*object.links[i]);
    this->links.push_back(new_link);
  }
  this->index=object.index;
  this->parent_segment=object.parent_segment;
}

void COpendriveRoadNode::add_link(COpendriveRoadNode *node,opendrive_mark_t mark)
{
  TOpendriveWorldPoint lane_end,node_start;
  double start_curvature,end_curvature,length;
  COpendriveLink *new_link;
  bool lane_found=false;

  for(unsigned int i=0;i<this->links.size();i++)
    if(this->links[i]->prev==this && this->links[i]->next==node)
      return;
  new_link=new COpendriveLink();
  new_link->set_prev(this);
  new_link->set_next(node);
  new_link->set_resolution(this->resolution);
  new_link->set_scale_factor(this->scale_factor);
  new_link->set_road_mark(mark);
  // get the curvature
  node_start=node->get_start_pose();
  for(unsigned int i=0;i<this->lanes.size();i++)
  {
    if(this->lanes[i]->get_id()<0)
      lane_end=this->lanes[i]->get_end_point();
    else
      lane_end=this->lanes[i]->get_start_point();
    if(fabs(lane_end.x-node_start.x)<this->resolution && fabs(lane_end.y-node_start.y)<this->resolution)
    {
      this->geometries[i]->get_curvature(start_curvature,end_curvature);
      length=this->geometries[i]->get_length();
      if(this->lanes[i]->get_id()>0)
      {
        start_curvature*=-1.0;
        end_curvature*=-1.0;
      }
      lane_found=true;
      break;
    }
  }
  if(!lane_found)
  {
    start_curvature=0.0;
    end_curvature=0.0;
    length=sqrt(pow(this->start_point.x-node_start.x,2.0)+pow(this->start_point.y-node_start.y,2.0));
  }
  new_link->generate(start_curvature,end_curvature,length,lane_found);
  this->links.push_back(new_link);
  node->links.push_back(new_link);
}

void COpendriveRoadNode::set_resolution(double res)
{
  this->resolution=res;

  for(unsigned int i=0;i<this->links.size();i++)
    this->links[i]->set_resolution(res);
}

void COpendriveRoadNode::set_scale_factor(double scale)
{
  this->scale_factor=scale;

  for(unsigned int i=0;i<this->links.size();i++)
    this->links[i]->set_scale_factor(scale);
}

void COpendriveRoadNode::set_index(unsigned int index)
{
  this->index=index;
}

void COpendriveRoadNode::set_parent_segment(COpendriveRoadSegment *segment)
{
  this->parent_segment=segment;
}

void COpendriveRoadNode::add_lane(const planView::geometry_type &geom_info,COpendriveLane *lane)
{
  COpendriveGeometry *geometry;

  // import geometry
  if(geom_info.line().present())
    geometry=new COpendriveLine();
  else if(geom_info.spiral().present())
    geometry=new COpendriveSpiral();
  else if(geom_info.arc().present())
    geometry=new COpendriveArc();
  else if(geom_info.paramPoly3().present())
    geometry=new COpendriveParamPoly3();
  else
    throw CException(_HERE_,"Unsupported or Missing geometry");
  geometry->set_scale_factor(this->scale_factor);
  geometry->load(geom_info);
  this->geometries.push_back(geometry);
  this->lanes.push_back(lane);
}

void COpendriveRoadNode::load(const planView::geometry_type &geom_info,COpendriveLane *lane)
{
  TOpendriveTrackPoint track_point;
  COpendriveGeometry *geometry;

  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    delete this->geometries[i];
    this->geometries[i]=NULL;
  }
  this->geometries.clear();
  // import geometry
  if(geom_info.line().present())
    geometry=new COpendriveLine();
  else if(geom_info.spiral().present())
    geometry=new COpendriveSpiral();
  else if(geom_info.arc().present())
    geometry=new COpendriveArc();
  else if(geom_info.paramPoly3().present())
    geometry=new COpendriveParamPoly3();
  else
    throw CException(_HERE_,"Unsupported or Missing geometry");
  geometry->set_scale_factor(this->scale_factor);
  geometry->load(geom_info);
  // import start position
  if(lane->get_id()<0)
  {
    track_point.t=lane->get_offset()-lane->get_width()/2.0;
    track_point.s=0.0;
  }
  else
  {
    track_point.t=lane->get_offset()+lane->get_width()/2.0;
    track_point.s=geometry->get_length();
  }
  track_point.heading=0.0;
  if(!geometry->get_world_pose(track_point,this->start_point))
  {
    delete geometry;
    geometry=NULL;
    throw CException(_HERE_,"Impossible to get world coordinates");
  }
  else
  {
    if(lane->get_id()>0)
      this->start_point.heading=normalize_angle(this->start_point.heading+3.14159);
  }
  this->geometries.push_back(geometry);
  for(unsigned int i=0;i<this->links.size();i++)
  {
    delete this->links[i];
    this->links[i]=NULL;
  }
  this->links.clear();
  this->lanes.clear();
  this->lanes.push_back(lane);
}

void COpendriveRoadNode::update_references(std::map<COpendriveRoadNode *,COpendriveRoadNode *> refs)
{
  for(unsigned int i=0;i<this->links.size();i++)
    this->links[i]->update_references(refs);
}

void COpendriveRoadNode::update_references(std::map<COpendriveRoadSegment *,COpendriveRoadSegment *> refs)
{
  this->parent_segment=refs[this->parent_segment];
}

void COpendriveRoadNode::update_references(std::map<COpendriveLane *,COpendriveLane *> refs)
{
  for(unsigned int i=0;i<this->lanes.size();i++)
    this->lanes[i]=refs[this->lanes[i]];
}

double COpendriveRoadNode::get_resolution(void) const
{
  return this->resolution;
}

double COpendriveRoadNode::get_scale_factor(void) const
{
  return this->scale_factor;
}

unsigned int COpendriveRoadNode::get_index(void) const
{
 return this->index;
}

const COpendriveRoadSegment &COpendriveRoadNode::get_parent_segment(void) const
{
  return *this->parent_segment;
}

TOpendriveWorldPoint COpendriveRoadNode::get_start_pose(void) const
{
  return this->start_point;
}

unsigned int COpendriveRoadNode::get_num_links(void) const
{
  return this->links.size();
}

const COpendriveLink &COpendriveRoadNode::get_link(unsigned int index) const
{
  if(index>=0 && index<this->links.size())
    return *this->links[index];
  else
    throw CException(_HERE_,"Invalid link index");
}

const COpendriveLink &COpendriveRoadNode::get_link_ending_at(unsigned int node_index) const
{
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(this->links[i]->get_next().get_index()==node_index)
      return *this->links[i];
  }

  throw CException(_HERE_,"No link in this node ends at the desired node");
}

unsigned int COpendriveRoadNode::get_num_lanes(void) const
{
  return this->lanes.size();
}

const COpendriveLane &COpendriveRoadNode::get_lane(unsigned int index) const
{
  if(index>=0 && index<this->lanes.size())
    return *this->lanes[index];
  else
    throw CException(_HERE_,"Invalid lane index");
}

unsigned int COpendriveRoadNode::get_num_geometries(void) const
{
  return this->geometries.size();
}

const COpendriveGeometry &COpendriveRoadNode::get_geometry(unsigned int index) const
{
  if(index>=0 && index<this->geometries.size())
    return *this->geometries[index];
  else
    throw CException(_HERE_,"Invalid geometry index");
}

unsigned int COpendriveRoadNode::get_closest_link(double x, double y,double heading,double angle_tol)const 
{
  double dist,min_dist=std::numeric_limits<double>::max();
  TOpendriveWorldPoint target;
  unsigned int closest_index;
  double angle;
  TPoint point;

  target.x=x;
  target.y=y;
  target.heading=heading;
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      this->links[i]->find_closest_world_point(target,point);
      angle=diff_angle(normalize_angle(heading),point.heading);
      if(fabs(angle)<angle_tol)
      {
        dist=sqrt(pow(point.x-x,2.0)+pow(point.y-y,2.0));
        if(dist<min_dist)
        {
          min_dist=dist;
          closest_index=i;
        }
      }
    }
  }

  return closest_index;
}

double COpendriveRoadNode::get_closest_point(double x, double y,double heading,TOpendriveWorldPoint &closest_point,double angle_tol) const
{
  double dist,min_dist=std::numeric_limits<double>::max();
  double angle;
  double length,closest_length;
  TOpendriveWorldPoint target;
  TPoint point;

  target.x=x;
  target.y=y;
  target.heading=heading;
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      length=this->links[i]->find_closest_world_point(target,point);
      angle=diff_angle(normalize_angle(heading),point.heading);
      if(fabs(angle)<angle_tol)
      {
        dist=sqrt(pow(point.x-x,2.0)+pow(point.y-y,2.0));
        if(dist<min_dist)
        {
          min_dist=dist;
          closest_point.x=point.x;
          closest_point.y=point.y;
          closest_point.heading=point.heading;
          closest_length=length;
        }
      }
    }
  }

  return closest_length;
}

void COpendriveRoadNode::get_closest_points(double x, double y,double heading,std::vector<TOpendriveWorldPoint> &closest_points,std::vector<double> &closest_lengths,double dist_tol,double angle_tol) const
{
  double dist;
  double angle;
  double length;
  TOpendriveWorldPoint target,new_point;
  TPoint point;

  target.x=x;
  target.y=y;
  target.heading=heading;
  closest_points.clear();
  closest_lengths.clear();
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      length=this->links[i]->find_closest_world_point(target,point);
      angle=diff_angle(normalize_angle(heading),point.heading);
      if(fabs(angle)<angle_tol)
      {
        dist=sqrt(pow(point.x-x,2.0)+pow(point.y-y,2.0));
        if(dist<=dist_tol)
        {
          new_point.x=point.x;
          new_point.y=point.y;
          new_point.heading=point.heading;
          closest_points.push_back(new_point);
          closest_lengths.push_back(length);
        }
      }
    }
  }
}

std::ostream& operator<<(std::ostream& out, COpendriveRoadNode &node)
{
  out << "    Node: " << node.get_index() << std::endl;
  out << "      start point: x: " << node.start_point.x << " y: " << node.start_point.y << " heading: " << node.start_point.heading << std::endl;
  out << "      Number of parent lanes: " << node.lanes.size() << std::endl;
  for(unsigned int i=0;i<node.lanes.size();i++)
  {
    out << "      Lane " << i << ":" << std::endl;
    out << "        lane id: " << node.lanes[i]->get_id() << std::endl;
    out << *node.geometries[i];
  }
  out << "      Number of links: " << node.links.size() << std::endl;
  for(unsigned int i=0;i<node.links.size();i++)
  {
    out << "      Link " << i << ":" << std::endl;
    out << *node.links[i];
  }
  return out;
}

COpendriveRoadNode::~COpendriveRoadNode()
{
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->start_point.x=0.0;
  this->start_point.y=0.0;
  this->start_point.heading=0.0;
  this->lanes.clear();
  for(unsigned int i=0;i<this->geometries.size();i++)
  {
    delete this->geometries[i];
    this->geometries[i]=NULL;
  }
  this->geometries.clear();
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(this->links[i]->prev==this)
    {
      delete this->links[i];
      this->links[i]=NULL;
    }
  }
  this->links.clear();
}

